(** PCP Solver *)
module PCPParser = struct

open Parser

let int = many digit

let int_list =
 between 
  (char '[' >> spaces) 
  (sep_by (spaces >> char ';' >> spaces) int)
  (spaces >> char ']')

let parse = 
 spaces >>
 int_list >>= fun d1 ->
 spaces >>
 int_list >>= fun d2 ->
 spaces >>
 return (Lst.zip d1 d2)

let parse input = match Parser.parse parse (Strng.of_string input) with
 | None ->  
  Format.printf "shape of dictionaries: \"[0;01;110] [100;00;11]\"\n";
  failwith ("parsing error on " ^ input)
 | Some ds -> ds
end

(*samples*)
(*shortest solution [0;1]*)
let pcp1 = PCPParser.parse "[01;0] [0;10]";; 

(*no solution*)
let pcp2 = PCPParser.parse "[01;11] [0;10]";;

(*shortest solution of length 76*)
let pcp3 = PCPParser.parse "[01;01;0;100] [0;110;100;10]";;

(*solution of length 132*)
let pcp4 = PCPParser.parse "[101;0;10;11] [1;01;110;10]";;

(*no solution*)
let pcp5 = PCPParser.parse "[01;00] [0;10]";;

(*solving*)
(* trim : ('a list * 'a list) -> ('a list * 'a list) *)
let rec trim = function
 | (x::xs,y::ys) when x = y -> trim (xs,ys)
 | d -> d

(* extend : ('a list * 'a list) -> ('a list * 'a list) *)
let extend (w1,w2) (d1,d2) = trim (w1@d1,w2@d2)

(* solve : ('a list * 'a list) list -> bool *)
let solve ds =
 let rec solve = function
  | [] -> false
  | ([],[])::_ -> true
  | (x::_,y::_)::ws when x <> y -> solve ws
  | w::ws -> solve (ws@Lst.map (extend w) ds)
 in solve (Lst.map trim ds)

